@@ -17,6 +17,8 @@ sub make_response {
     my $page = <<'HTML';
 <!doctype html><html><body>
 <link rel="stylesheet" href="/stylesheet.css">
+<link rel="stylesheet" href="/stylesheet.css">
+<link rel="icon" href="/favicon.ico" type="image/vnd.microsoft.icon">
 <h1>HTTP standard conformance tests - HTTP evader</h1>
 
 <p>
@@ -117,7 +119,9 @@ innocent image payload</a></p>
 
 <p>
 This bulk test will use "iframe" tags to download an innocent HTML to check which
-uncommon responses can be used to load iframes.
+uncommon responses can be used to load iframes. <b>Warning!</b>: IE and Edge seem
+to have serious problems with some test cases here and will render the page
+unresponsive.
 </p>
 <p id=test_iframe class=runtest><a href="/autohtml/all/parent_set_success.html">Run Test with
 innocent iframe payload</a></p>
@@ -218,10 +222,10 @@ sub auto_xhr {
 	for($_->TESTS) {
 	    if ($isbad) {
 		$html .= sprintf("checks.push({ page:'%s', desc:'%s', valid:%d, harmless_page: '%s', file: '%s'  });\n",
-		    $_->url($page), quotemeta(html_escape($_->DESCRIPTION)), $_->VALID, $_->url('novirus.txt'),'novirus.txt')
+		    url_encode($_->url($page)), quotemeta(html_escape($_->DESCRIPTION)), $_->VALID, url_encode($_->url('novirus.txt')),'novirus.txt')
 	    } else {
 		$html .= sprintf("checks.push({ page:'%s', desc:'%s', valid:%d, file:'%s' });\n",
-		    $_->url($page), quotemeta(html_escape($_->DESCRIPTION)), $_->VALID,$page)
+		    url_encode($_->url($page)), quotemeta(html_escape($_->DESCRIPTION)), $_->VALID,$page)
 	    }
 	}
 
@@ -231,6 +235,7 @@ sub auto_xhr {
     return "HTTP/1.0 200 ok\r\n".
 	"Content-type: text/html\r\n".
 	"Content-length: ".length($html)."\r\n".
+	"ETag: ".App::DubiousHTTP->VERSION."\r\n".
 	"\r\n".
 	$html;
 }
@@ -281,7 +286,7 @@ sub _auto_imgjshtml {
 	next if $cat ne 'all' && $_->ID ne $cat;
 	for($_->TESTS) {
 	    my $xid = quotemeta(html_escape($_->LONG_ID));
-	    my $url = $_->url($page);
+	    my $url = url_encode($_->url($page));
 	    my $html = $mkhtml->("$url?rand=$rand",$xid);
 	    $jsglob .= "checks.push({ "
 		. "page: '$url', xid: '$xid', "
@@ -298,6 +303,7 @@ sub _auto_imgjshtml {
     return "HTTP/1.0 200 ok\r\n".
 	"Content-type: text/html\r\n".
 	"Content-length: ".length($html)."\r\n".
+	"ETag: ".App::DubiousHTTP->VERSION."\r\n".
 	"\r\n".
 	$html;
 }
@@ -334,7 +340,7 @@ You need to have JavaScript enabled to run this tests.
 <div id=vendor_notice> </div>
 <div id=nobad> </div>
 <div id=urlblock> </div>
-<div id=evasions></td>
+<div id=evasions></div>
 <div id=process></div>
 <div id=evadable> </div>
 <div id=overblock> </div>
@@ -713,9 +719,6 @@ function check_nonxhr_result(xid,type,val) {
     if (current_test && current_test['xid'] == xid) {
 	window.clearTimeout(nonxhr_timer);
 	test = current_test;
-	if (!val && !test['status'] && event && event.type == 'error') {
-	    val = 'fail';
-	}
 	if (val) {
 	    add_debug( val + ": " + test['desc'], test);
 	    test['status'] = val;
@@ -726,7 +729,6 @@ function check_nonxhr_result(xid,type,val) {
 	    }
 	} else if (!test['status']) {
 	    // no final result, wait some more time
-	    // _log(event);
 	    _log("defer " + xid);
 	    open_checks[xid] = test;
 	}
@@ -759,9 +761,6 @@ function check_nonxhr_result(xid,type,val) {
 
     if (open_checks[xid]) {
 	test = open_checks[xid];
-	if (!val && !test['status'] && event && event.type == 'error') {
-	    val = 'fail';
-	}
 	if (val) {
 	    delete open_checks[xid];
 	    add_debug( "delayed " + val + ": " + test['desc'], test);
@@ -989,5 +988,22 @@ HTML
     }
 }
 
+sub manifest {
+    my ($self,$cat,$page,$spec) = @_;
+    my $data = "trivial | /clen/$page/close,clen,content | 3 | trivial response for retrieving body\n";
+    for(@cat) {
+	next if $cat ne 'all' && $_->ID ne $cat;
+	for($_->TESTS) {
+	    $data .= sprintf("%s | %s | %s | %s\n",
+		$_->LONG_ID, $_->url($page), $_->VALID, $_->DESCRIPTION);
+	}
+    }
+    return "HTTP/1.0 200 ok\r\n".
+	"Content-type: text/plain\r\n".
+	"Content-length: ".length($data)."\r\n".
+	"\r\n".
+	$data;
+}
+
 
 1;
