@@ -2,6 +2,7 @@ use strict;
 use warnings;
 package App::DubiousHTTP::Tests::Chunked;
 use App::DubiousHTTP::Tests::Common;
+use Compress::Raw::Zlib;
 
 SETUP(
     'chunked',
@@ -32,15 +33,36 @@ DESC
     [ INVALID, 'size-cr' => "size followed by <CR>" ],
     [ INVALID, 'size-lf' => "size followed by <LF>" ],
     [ INVALID, 'size-x' => "size followed by char 'x'" ],
+    [ INVALID, 'size-\054' => "size followed by char ','" ],
+    [ INVALID, 'size-\000' => "size followed by char \\000" ],
+    [ INVALID, 'size-\013' => "size followed by char \\v" ],
+    [ INVALID, 'size-\014' => "size followed by char \\f" ],
     [ INVALID, 'size-spacex' => "size followed by space and char 'x'" ],
     [ INVALID, 'space-size' => "size prefixed by space" ],
     [ INVALID, 'tab-size' => "size prefixed by tab" ],
     [ INVALID, 'cr-size' => "size prefixed by cr" ],
+    [ INVALID, 'lf-size' => "size prefixed by lf" ],
+    [ INVALID, 'crlf-size' => 'size prefixed by "\r\n"' ],
+    [ INVALID, 'crlf-crlf-size' => 'size prefixed by "\r\n\r\n"' ],
+    [ INVALID, 'crlf-x-crlf-size' => 'size prefixed by "\r\nx\r\n"' ],
     [ INVALID, 'x-size' => "size prefixed by char 'x'" ],
+    [ INVALID, '\054-size' => "size prefixed by char ','" ],
+    [ INVALID, '\073-size' => "size prefixed by char ';'" ],
+    [ INVALID, '\000-size' => "size prefixed by char \\000" ],
+    [ INVALID, '\013-size' => "size prefixed by char \\v" ],
+    [ INVALID, '\014-size' => "size prefixed by char \\f" ],
     [ INVALID, 'xspace-size' => "size prefixed by char 'x' and space" ],
     [ UNCOMMON_VALID, 'final=00' => 'final chunk size "00"' ],
     [ UNCOMMON_VALID, 'final=00000000000000000000' => 'final chunk size "00000000000000000000"' ],
     [ INVALID, 'final=0x' => 'final chunk size "0x"' ],
+    [ INVALID, 'finalchunk=0\012' => 'final chunk just "0\n"' ],
+    [ INVALID, 'finalchunk=0\015' => 'final chunk just "0\r"' ],
+    [ INVALID, 'finalchunk=0' => 'final chunk just "0"' ],
+    [ INVALID, 'finalchunk=0\012\012' => 'final chunk "0\n\n"' ],
+    [ INVALID, 'finalchunk=0\012\040\012' => 'final chunk "0\n<space>\n"' ],
+    [ INVALID, 'finalchunk=0\012\015\012' => 'final chunk "0\n\r\n"' ],
+    [ INVALID, 'finalchunk=0\012\015\015\012' => 'final chunk "0\n\r\r\n"' ],
+    [ INVALID, 'finalchunk=0\015\012foobar\015\012' => 'final chunk "0\r\nfoobar\r\n"' ],
 
     [ 'VALID: (but uncommon) use of extensions in chunked header' ],
     [ UNCOMMON_VALID, 'chunk-ext-junk' => "chunked with some junk chunk extension" ],
@@ -54,16 +76,23 @@ DESC
     # safari does not like it, so mark it as uncommon
     [ UNCOMMON_VALID, 'chunked,clen200' => 'chunking and content-length header with double length, served chunked'],
     [ UNCOMMON_VALID, 'chunked,clen50'  => 'chunking and content-length header with half length, served chunked'],
+    [ INVALID, 'addjunk,chunked,clen50'  => 'content+junk, chunked, content-length header includes content only' ],
 
     [ 'INVALID: chunking is only allowed with HTTP/1.1' ],
     [ INVALID, 'chunked,http10' => 'Chunked Header and HTTP/1.0. Served chunked.'],
     [ INVALID, 'chunked,clen,http10' => 'Chunked Header and Content-length and HTTP/1.0. Served chunked.'],
     [ INVALID, 'clen,chunked,http10' => 'Content-length Header and Chunked and HTTP/1.0. Served chunked.'],
+    [ INVALID, 'chunked,http10,gzip' => 'Chunked Header and HTTP/1.0. Served chunked with gzip.'],
+    [ INVALID, 'chunked,clen,http10,gzip' => 'Chunked Header and Content-length and HTTP/1.0. Served chunked with gzip.'],
+    [ INVALID, 'clen,chunked,http10,gzip' => 'Content-length Header and Chunked and HTTP/1.0. Served chunked with gzip.'],
 
     [ 'VALID: chunked header should be ignored with HTTP/1.0' ],
     [ UNCOMMON_VALID, 'chunked,http10,do_clen' => 'Chunked Header and HTTP/1.0. Not served chunked.'],
     [ UNCOMMON_VALID, 'chunked,clen,http10,do_clen' => 'Chunked Header and Content-length and HTTP/1.0. Not served chunked.'],
     [ UNCOMMON_VALID, 'clen,chunked,http10,do_clen' => 'Content-length Header and Chunked and HTTP/1.0. Not served chunked.'],
+    [ UNCOMMON_VALID, 'chunked,http10,do_clen,gzip' => 'Chunked Header and HTTP/1.0. Not served chunked. Compressed with gzip.'],
+    [ UNCOMMON_VALID, 'chunked,clen,http10,do_clen,gzip' => 'Chunked Header and Content-length and HTTP/1.0. Not served chunked. Compressed with gzip.'],
+    [ UNCOMMON_VALID, 'clen,chunked,http10,do_clen,gzip' => 'Content-length Header and Chunked and HTTP/1.0. Not served chunked. Compressed with gzip.'],
 
     [ 'INVALID: chunking with invalid HTTP versions' ],
     [ INVALID, 'chunked,HTTP/1.2' => 'Chunked Header and HTTP/1.2. Served chunked.'],
@@ -115,9 +144,11 @@ DESC
     [ INVALID,'chunkedcr-,do_clen' => "Transfer-Encoding:chunked<CR><space>. Not served chunked."],
     [ INVALID,'ce-chunked,do_chunked' => "Content-encoding chunked instead of Transfer-encoding. Served chunked."],
 
-    [ 'INVALID: hiding with another Transfer-Encoding' ],
+    [ 'INVALID: hiding with another Transfer-Encoding header' ],
     [ INVALID, 'xte,chunked,do_chunked' => "double Transfer-Encoding: first junk, last chunked. Served chunked." ],
+    [ INVALID, 'xte,chunked,do_chunked,gzip' => "double Transfer-Encoding: first junk, last chunked. Served chunked and gzipped." ],
     [ INVALID, 'chunked,xte,do_chunked' => "double Transfer-Encoding: first chunked, last junk. Served chunked." ],
+    [ INVALID, 'chunked,xte,do_chunked,gzip' => "double Transfer-Encoding: first chunked, last junk. Served chunked and gzipped." ],
     [ INVALID, 'xte,chunked,xte,do_chunked' => "triple Transfer-Encoding: first junk, then chunked, then junk again. Served chunked." ],
     [ INVALID, 'xte,chunked,do_clen' => "double Transfer-Encoding: first junk, last chunked. Not served chunked." ],
     [ INVALID, 'chunked,xte,do_clen' => "double Transfer-Encoding: first chunked, last junk. Not served chunked." ],
@@ -125,8 +156,11 @@ DESC
     [ INVALID, 'xte,chunked,clen,do_chunked' => "double Transfer-Encoding: first junk, last chunked. Also Content-length header. Served chunked." ],
     [ INVALID, 'xte,chunked,xte,clen,do_chunked' => "triple Transfer-Encoding: first junk, then chunked, then junk again. Also Content-length header. Served chunked." ],
     [ INVALID, 'chunked,xte,clen,do_clen' => "double Transfer-Encoding: first chunked, last junk. Also Content-length header. Not served chunked." ],
+    [ INVALID, 'chunked,xte,clen,do_clen,gzip' => "double Transfer-Encoding: first chunked, last junk. Also Content-length header. Not served chunked. Compressed with gzip." ],
     [ INVALID, 'xte,chunked,clen,do_clen' => "double Transfer-Encoding: first junk, last chunked. Also Content-length header. Not served chunked." ],
+    [ INVALID, 'xte,chunked,clen,do_clen,gzip' => "double Transfer-Encoding: first junk, last chunked. Also Content-length header. Not served chunked. Compressed with gzip." ],
     [ INVALID, 'chunked,clen,do_clen' => 'chunking and content-length, not served chunked'],
+    [ INVALID, 'chunked,clen,do_clen,gzip' => 'chunking and content-length, not served chunked. Compressed with gzip.'],
 
     [ 'INVALID: hiding the Transfer-Encoding header' ],
     [ INVALID, 'space-colon-chunked,do_chunked' => '"Transfer-Encoding<space>:", served chunked' ],
@@ -138,6 +172,7 @@ DESC
     [ INVALID, 'colon-colon-chunked,do_chunked' => '"Transfer-Encoding::", served chunked' ],
     [ UNCOMMON_INVALID, 'colon-colon-chunked,do_clen' => '"Transfer-Encoding::", not served chunked' ],
     [ INVALID, 'cronly-chunked,do_chunked' => 'Transfer-Encoding with only <CR> as line delimiter before, served chunked' ],
+    [ INVALID, 'crxonly-chunked,do_chunked' => 'Only <CR> as line delimiter followed by "xTransfer-Encoding", served chunked' ],
     [ UNCOMMON_INVALID, 'cronly-chunked,do_clen' => 'Transfer-Encoding with only <CR> as line delimiter before, not served chunked' ],
     [ UNCOMMON_INVALID, 'lfonly-chunked,do_chunked' => 'Transfer-Encoding with only <LF> as line delimiter before, served chunked' ],
     [ INVALID, 'lfonly-chunked,do_clen' => 'Transfer-Encoding with only <LF> as line delimiter before, not served chunked' ],
@@ -162,10 +197,11 @@ sub make_response {
     return make_index_page() if $page eq '';
     my ($hdr,$data) = content($page,$self->ID."-".$spec) or die "unknown page $page";
     my $version = 'HTTP/1.1';
-    my ($te,@chunks,%chunkmod);
+    my ($te,@chunks,%chunkmod,$clen);
     my $sizefmt = '%x';
     my $before_chunks = '';
     my $final = '0';
+    my $finalchunk;
     for (split(',',$spec)) {
 	if ( m{^(x|-|nl|lf|cr)*chunked(x|-|nl|lf|cr)*$}i ) {
 	    s{-}{ }g;
@@ -181,11 +217,12 @@ sub make_response {
 	    $c =~s{cr}{\r}g;
 	    $te = 'chunked';
 	    $hdr .= "Connection: close\r\nTransfer-Encoding$c: chunked\r\n"
-	} elsif ( my ($crlf) = m {^(lf|cr)only-chunked$} ) {
+	} elsif ( my ($crlf) = m {^((?:lf|cr|x)+)only-chunked$} ) {
 	    $te = 'chunked';
 	    $hdr = "X-Foo: bar" if $hdr !~s{\r\n\z}{};
-	    $hdr .= ($crlf eq 'lf') ? "\n":"\r";
-	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    $crlf =~s{lf}{\n}g;
+	    $crlf =~s{cr}{\r}g;
+	    $hdr .= $crlf . "Transfer-Encoding: chunked\r\nConnection: close\r\n";
 	} elsif ( $_ eq '1chunk' ) {
 	    $hdr .= "Transfer-Encoding: chunked\r\n";
 	    @chunks = $data;
@@ -194,7 +231,7 @@ sub make_response {
 	} elsif ( $_ eq 'ce-chunked' ) {
 	    $hdr .= "Content-Encoding: chunked\r\nConnection: close\r\n"
 	} elsif ( $_ =~ m{^clen(\d+)?$} ) {
-	    $hdr .= "Content-length: ". int(($1||100)/100*length($data)) ."\r\n"
+	    $clen = $1 || 100;
 	} elsif ( $_ eq 'http10' ) {
 	    $version = "HTTP/1.0";
 	} elsif ( $_ =~m{^HTTP/\S+}i ) {
@@ -217,7 +254,7 @@ sub make_response {
 	    $eol =~s{cr}{\r}g;
 	    $eol =~s{lf}{\n}g;
 	    $chunkmod{lineend} = $eol;
-	} elsif ( m{^(-|space|cr|lf|tab|x)*(0*)(uc)?size(-|space|cr|lf|tab|x)*$}) {
+	} elsif ( m{^(-|space|cr|lf|tab|x|\\[0-7]{3})*(0*)(uc)?size(-|space|cr|lf|tab|x|\\[0-7]{3})*$}) {
 	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
 	    @chunks = ( $data =~m{(.{1,15})}smg,'') if ! @chunks;
 	    s{ucsize}{%X};
@@ -229,6 +266,7 @@ sub make_response {
 	    s{tab}{\t}g;
 	    s{cr}{\r}g;
 	    s{lf}{\n}g;
+	    s{\\([0-7]{3})}{ chr(oct($1)) }eg;
 	    $sizefmt = $_;
 	} elsif (m{((?:space|tab|cr|lf)*)-before-chunks}) {
 	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
@@ -240,10 +278,21 @@ sub make_response {
 	} elsif (m{^final=(.*)$}) {
 	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
 	    $final = $1;
+	} elsif (m{^finalchunk=(.*)$}) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    (my $d = $1 ) =~ s{\\([0-7]{3})}{ chr(oct($1)) }eg;
+	    $finalchunk = $d;
+	} elsif ( $_ eq 'addjunk' ) {
+	    $data .= "x" x length($data);
+	} elsif ( $_ eq 'gzip' ) {
+	    $data = _compress($data,'gzip');
+	    $hdr .= "Content-Encoding: gzip\r\n";
 	} else {
 	    die $_
 	}
     }
+    $hdr .= "Content-length: ". int($clen/100*length($data)) ."\r\n" 
+	if defined $clen;
     $hdr = "$version 200 ok\r\n$hdr";
     $te ||= $hdr =~m{^Transfer-Encoding:}im ? 'chunked':'clen';
     @chunks = ( $data =~m{(.{1,5})}smg,'') if $te eq 'chunked' && ! @chunks;
@@ -264,6 +313,7 @@ sub make_response {
 	    $end = sprintf("%x%s%s%s",$last->[0]+10,$last->[2]||'',$nl,$last->[1]);
 	}
 
+	$finalchunk //= "$final$nl$nl";
 	$data = join("",map { 
 	    $_->[0] ? sprintf("$sizefmt%s%s%s%s",
 		$_->[0],          # size
@@ -271,11 +321,22 @@ sub make_response {
 		$nl,
 		$_->[1],
 		$nl
-	    ) : "$final$nl$nl"
+	    ) : $finalchunk
 	} @chunks).$end;
     }
     return "$hdr\r\n$before_chunks$data";
 }
 
+sub _compress {
+    my ($data,$w) = @_;
+    my $zlib = Compress::Raw::Zlib::Deflate->new(
+	-WindowBits => $w eq 'gzip' ? WANT_GZIP : -MAX_WBITS(),
+	-AppendOutput => 1,
+    );
+    my $newdata = '';
+    $zlib->deflate( $data, $newdata);
+    $zlib->flush($newdata,Z_FINISH);
+    return $newdata;
+}
 
 1;
