@@ -27,19 +27,36 @@ DESC
     # these should be fine
     [ 'VALID: correct compressed requests' ],
     [ SHOULDBE_VALID, 'ce:gzip;gzip' => 'content-encoding gzip, served gzipped'],
-    [ VALID, 'ce:gzip;gzip2p' => 'content-encoding gzip, served gzipped with 2 compressed blocks'],
     [ UNCOMMON_VALID, 'ce:x-gzip;gzip' => 'content-encoding "x-gzip", served gzipped'], # not IE11
     [ SHOULDBE_VALID, 'ce:deflate;deflate' => 'content-encoding deflate, served with deflate'],
-    [ VALID, 'ce:deflate;deflate2p' => 'content-encoding deflate, served with deflate with 2 compressed blocks'],
     [ VALID, 'ce:deFLaTe;deflate' => 'content-encoding deflate mixed case, served with deflate'],
 
+    # various kinds of flush between compression parts
+    [ 'UNCOMMON_VALID: various kinds of flush between compression parts' ],
+    [ UNCOMMON_VALID, 'ce:gzip;gzip2p,partial' => 'content-encoding gzip, served gzipped with 2 compressed blocks with partial flush in between'],
+    [ UNCOMMON_VALID, 'ce:deflate;deflate2p,partial' => 'content-encoding deflate, served with deflate with 2 compressed blocks with partial flush in between'],
+    [ UNCOMMON_VALID, 'ce:gzip;gzip2p,block' => 'content-encoding gzip, served gzipped with 2 compressed blocks with block flush in between'],
+    [ UNCOMMON_VALID, 'ce:deflate;deflate2p,block' => 'content-encoding deflate, served with deflate with 2 compressed blocks with block flush in between'],
+    [ UNCOMMON_VALID, 'ce:gzip;gzip2p,sync' => 'content-encoding gzip, served gzipped with 2 compressed blocks with sync flush in between'],
+    [ UNCOMMON_VALID, 'ce:deflate;deflate2p,sync' => 'content-encoding deflate, served with deflate with 2 compressed blocks with sync flush in between'],
+    [ UNCOMMON_VALID, 'ce:gzip;gzip2p' => 'content-encoding gzip, served gzipped with 2 compressed blocks with full flush in between'],
+    [ UNCOMMON_VALID, 'ce:deflate;deflate2p' => 'content-encoding deflate, served with deflate with 2 compressed blocks with full flush in between'],
+    [ INVALID, 'ce:gzip;gzip2p,finish' => 'content-encoding gzip, served gzipped with 2 compressed blocks with finish in between'],
+    [ INVALID, 'ce:deflate;deflate2p,finish' => 'content-encoding deflate, served with deflate with 2 compressed blocks with finish in between'],
+
     [ 'VALID: lzma (supported by at least Opera)' ],
     [ UNCOMMON_VALID, 'ce:lzma;lzma1' => 'content-encoding lzma, lzma1 (lzma_alone) encoded'],
 
+    [ 'VALID: brotli (supported by at least Firefox when used with https)' ],
+    [ UNCOMMON_VALID, 'ce:br;brotli' => 'content-encoding br, encoded with brotli'],
+
+    [ 'INVALID: gzip header combined with zlib (RFC1952) instead of deflate (RFC1951)' ],
+    [ INVALID, 'ce:gzip;gzip-zlib' => 'content-encoding gzip, encoded with zlib prefixed by gzip header'],
+
     # these might be strange/unsupported
     [ 'VALID: less common but valid requests' ],
-    [ UNCOMMON_INVALID, 'ce:deflate;deflate-raw' => 'content-encoding deflate, served with RFC1950 style deflate (zlib)'],
-    [ UNCOMMON_INVALID, 'ce:deflate;deflate-raw2p' => 'content-encoding deflate, served with RFC1950 style deflate (zlib) with 2 compressed blocks'],
+    [ UNCOMMON_VALID, 'ce:deflate;zlib' => 'content-encoding deflate, served with RFC1950 style deflate (zlib)'],
+    [ UNCOMMON_VALID, 'ce:deflate;zlib2p' => 'content-encoding deflate, served with RFC1950 style deflate (zlib) with 2 compressed blocks'],
     [ UNCOMMON_VALID, 'ce:nl-gzip;gzip' => 'content-encoding gzip but with continuation line, served gzipped'],
     [ UNCOMMON_VALID, 'ce:nl-deflate;deflate' => 'content-encoding deflate but with continuation line, served with deflate'],
     [ UNCOMMON_VALID, 'ce:nl-nl-deflate;deflate' => 'content-encoding deflate but with double continuation line, served with deflate'],
@@ -119,14 +136,13 @@ DESC
     # and the bad ones
     [ 'INVALID: incorrect compressed response, should not succeed' ],
     [ INVALID, 'ce:x-deflate;deflate' => 'content-encoding x-deflate, served with deflate'],
-    [ INVALID, 'ce:x-deflate;deflate-raw' => 'content-encoding x-deflate, served with RFC1950 style deflate'],
+    [ INVALID, 'ce:x-deflate;zlib' => 'content-encoding x-deflate, served with RFC1950 style deflate (zlib)'],
     [ INVALID, 'ce:gzipx;gzip' => 'content-encoding "gzipx", served with gzip' ],
     [ INVALID, 'ce:xgzip;gzip' => 'content-encoding "xgzip", served with gzip' ],
     [ INVALID, 'ce:gzip_x;gzip' => 'content-encoding "gzip x", served with gzip' ],
     [ INVALID, 'ce:x_gzip;gzip' => 'content-encoding "x gzip", served with gzip' ],
     [ INVALID, 'ce:deflate;gzip' => 'content-encoding deflate but served with gzip'],
     [ INVALID, 'ce:gzip;deflate' => 'content-encoding gzip but served with decode'],
-    [ INVALID, 'ce:gzip;gzip-split' => 'content-encoding gzip, content split into 2 gzip parts concatenated'],
     [ INVALID, 'ce:deflate' => 'content-encoding "deflate", not encoded'],
     [ INVALID, 'ce:deflate,' => 'content-encoding "deflate,", not encoded'],
     [ INVALID, 'ce:deflate-nl-,' => 'content-encoding "deflate<nl> ,", not encoded'],
@@ -155,8 +171,10 @@ DESC
     [ UNCOMMON_INVALID, 'ce-colon-colon-gzip' => '"Content-Encoding:: gzip", served not with gzip' ],
 
     [ INVALID, 'cronly-deflate;deflate' => 'Content-Encoding with only <CR> as line delimiter before, served deflate' ],
+    [ INVALID, 'crxonly-deflate;deflate' => 'Only <CR> as line delimiter followed by "xContent-Encoding", served deflate' ],
     [ UNCOMMON_INVALID, 'cronly-deflate' => 'Content-Encoding with only <CR> as line delimiter before, not served deflate' ],
     [ INVALID, 'cronly-gzip;gzip' => 'Content-Encoding with only <CR> as line delimiter before, served gzip' ],
+    [ INVALID, 'crxonly-gzip;gzip' => 'Only <CR> as line delimiter followed by "xContent-Encoding", served gzip' ],
     [ UNCOMMON_INVALID, 'cronly-gzip' => 'Content-Encoding with only <CR> as line delimiter before, not served gzip' ],
 
     [ UNCOMMON_INVALID, 'lfonly-deflate;deflate' => 'Content-Encoding with only <LF> as line delimiter before, served deflate' ],
@@ -223,6 +241,9 @@ DESC
     [ INVALID,'ce-space-colon-gzip;gzip;replace:-4,1^ff', 'invalidate length (hide gzip with "content-encoding : gzip")'],
     [ INVALID,'ce-space-colon-gzip;gzip;replace:-4,4=', 'remove length (hide gzip with "content-encoding : gzip")'],
     [ INVALID,'ce-space-colon-gzip;gzip;replace:-8,8=', 'remove checksum and length (hide gzip with "content-encoding : gzip")'],
+
+    # data before gzip
+    [ INVALID,'ce:gzip;gzip;\012-before-body','new line at start of gzip body' ],
 );
 
 sub make_response {
@@ -231,8 +252,11 @@ sub make_response {
     my ($hdr,$data) = content($page,$self->ID."-".$spec) or die "unknown page $page";
     my $version = '1.1';
     my $clen_extend;
+    my $body_prefix = '';
     for (split(';',$spec)) {
-	if ( my ($field,$v) = m{^(ce|te):(.*)$}i ) {
+	if ($_ eq 'ce:rfc2047-deflate') {
+	    $hdr .= "Content-Encoding: =?UTF-8?B?ZGVmbGF0ZQo=?=\r\n";
+	} elsif ( my ($field,$v) = m{^(ce|te):(.*)$}i ) {
 	    my $changed;
 	    $changed++ if $v =~s{(?<=cr|lf|nl)-}{ }g;
 	    $changed++ if $v =~s{cr}{\r}g;
@@ -243,36 +267,34 @@ sub make_response {
 	    $hdr .= "Connection: close\r\n" if $changed;
 	    $hdr .= $field eq 'ce' ? 'Content-Encoding:':'Transfer-Encoding:';
 	    $hdr .= "$v\r\n";
-	} elsif ( m{^(?:(gzip)|deflate(-raw)?)(?:(\d+)p)?$} ) {
+	} elsif ( m{^(?:(gzip)|deflate|(zlib))(?:(\d+)p)?(?:,(sync|partial|block|full|finish))?$} ) {
 	    my $zlib = Compress::Raw::Zlib::Deflate->new(
 		-WindowBits => $1 ? WANT_GZIP : $2 ? +MAX_WBITS() : -MAX_WBITS(),
 		-AppendOutput => 1,
 	    );
 	    my $size = int(length($data)/($3||1)) || 1;
-	    my $newdata = '';
+	    my @chunks;
 	    while ($data ne '') {
-		my $out = '';
-		$zlib->deflate(substr($data,0,$size,''), $out);
-		$zlib->flush($out,Z_FULL_FLUSH);
-		$newdata .= $out;
+		push @chunks,substr($data,0,$size,'')
+	    }
+	    my $flush =
+		! $4 ? Z_FULL_FLUSH :
+		$4 eq 'partial' ? Z_PARTIAL_FLUSH :
+		$4 eq 'sync'    ? Z_SYNC_FLUSH :
+		$4 eq 'full'    ? Z_FULL_FLUSH :
+		$4 eq 'block'   ? Z_BLOCK :
+		$4 eq 'finish'  ? Z_FINISH :
+		die $4;
+	    my $newdata = '';
+	    while (@chunks) {
+		$zlib->deflate( shift(@chunks), $newdata);
+		if (defined $flush && @chunks) {
+		    $zlib->flush($newdata,$flush);
+		    $zlib->deflateReset if $flush == Z_FINISH;
+		}
 	    }
 	    $zlib->flush($newdata,Z_FINISH);
 	    $data = $newdata;
-	} elsif ( $_ =~m{^gzip-split(\d+)?$} ) {
-	    my $size = int(length($data)/($1||2)) || 1;
-	    my $newdata = '';
-	    while ($data ne '') {
-		my $zlib = Compress::Raw::Zlib::Deflate->new(
-		    -WindowBits => WANT_GZIP,
-		    -AppendOutput => 1,
-		);
-		my $out = '';
-		$zlib->deflate(substr($data,0,$size,''), $out);
-		$zlib->flush($out,Z_FINISH);
-		$newdata .= $out;
-	    }
-	    $data = $newdata;
-
 	} elsif (m{^(lzma[12]|xz)$}) {
 	    my ($lzma,$status) = 
 		$_ eq 'xz'    ? Compress::Raw::Lzma::EasyEncoder->new(AppendOutput => 1)  :
@@ -284,16 +306,32 @@ sub make_response {
 	    $lzma->flush($newdata,LZMA_FINISH) == LZMA_STREAM_END or die "failed to close lzma stream";
 	    $data = $newdata;
 
+	} elsif ($_ eq 'brotli') {
+	    $data = bro_compress($data) or do {
+		# no brotli for this content
+		return "HTTP/$version 500 no brotli\r\nContent-length:0\r\n\r\n";
+	    };
+	} elsif ($_ eq 'gzip-zlib') {
+	    my $zlib = Compress::Raw::Zlib::Deflate->new(
+		-WindowBits => +MAX_WBITS(),
+		-AppendOutput => 1,
+	    );
+	    my $gzip_hdr = pack("CCCCVCC",0x1f,0x8b,0x8,0,0,2,0);
+	    my $gzip_trailer = pack("VV",Compress::Raw::Zlib::crc32($data),length($data));
+	    my $newdata = '';
+	    $zlib->deflate($data,$newdata);
+	    $zlib->flush($newdata,Z_FINISH);
+	    $data = $gzip_hdr.$newdata.$gzip_trailer;
+
 	} elsif (m{^ce-space-colon-(.*)}) {
 	    $hdr .= "Content-Encoding : $1\r\n";
 	} elsif (m{^ce-colon-colon-(.*)}) {
 	    $hdr .= "Content-Encoding:: $1\r\n";
-	} elsif ( my ($crlf,$encoding) = m{^(lf|cr)only-(.*)}) {
+	} elsif ( my ($crlf,$encoding) = m{^((?:lf|cr|x)+)only-(.*)}) {
 	    $hdr = "X-Foo: bar" if $hdr !~s{\r\n\z}{};
-	    $hdr .= ($crlf eq 'lf') ? "\n":"\r";
-	    $hdr .= "Content-Encoding: $encoding\r\n";
-	} elsif ($_ eq 'ce:rfc2047-deflate') {
-	    $hdr .= "Content-Encoding: =?UTF-8?B?ZGVmbGF0ZQo=?=\r\n";
+	    $crlf =~s{cr}{\r}g;
+	    $crlf =~s{lf}{\n}g;
+	    $hdr .= $crlf . "Content-Encoding: $encoding\r\n";
 	} elsif ( my ($off,$len,$op,$replacement) = m{replace:(-?\d+),(\d+)([=|^])(.*)}) {
 	    $replacement = pack('C*',map { hex($_) } $replacement=~m{(..)}g);
 	    if ($op eq '=') {
@@ -310,11 +348,15 @@ sub make_response {
 	    $clen_extend = $1
 	} elsif ($_ eq 'noclen') {
 	    $clen_extend = -1;
+	} elsif (m{(.+)-before-body$}) {
+	    ( my $d = $1 ) =~s{\\([0-7]{3})}{ chr(oct($1)) }esg;
+	    $body_prefix .= $d;
 	} else {
 	    die $_
 	}
     }
 
+    $data = $body_prefix . $data;
     my $len = length($data);
     if (!$clen_extend) {
 	$hdr = "Content-length: ".length($data)."\r\n".$hdr;
